/* GSKrabManager.m - this file is part of GSKrab
 *
 * Copyright (C) 2006 Wolfgang Sourdeau
 *
 * Author: Wolfgang Sourdeau <Wolfgang@Contre.COM>
 *
 * This file is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This file is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.  If not, write to
 * the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#import <AppKit/NSWorkspace.h>
#import <Foundation/NSConnection.h>
#import <Foundation/NSDistantObject.h>
#import <Foundation/NSHost.h>
#import <Foundation/NSMapTable.h>
#import <Foundation/NSNotification.h>
#import <Foundation/NSPort.h>
#import <Foundation/NSPortNameServer.h>
#import <Foundation/NSString.h>
#import <Foundation/NSTask.h>
#import <Foundation/NSThread.h>
#import <Foundation/NSValue.h>

#import "GSKrabDesktopInterface.h"
#import "GSKrabServer.h"

#ifdef __WIN32__
#import "GSKrabWin32Interface.h"
#define DesktopInterface GSKrabWin32Interface
#else
#import "GSKrabX11Interface.h"
#define DesktopInterface GSKrabX11Interface
#endif

#import "GSKrabManager.h"

static NSNotificationCenter *nc = nil;

@implementation GSKrabManager

+ (void) initialize
{
  if (!nc)
    nc = [NSNotificationCenter defaultCenter];
}

// - (NSDictionary **) _knownKeys
// {
//   NSMutableDictionary **dicts;
//   NSArray *sampleKeys;

//   dicts = malloc (sizeof (NSDictionary *) * 2);
//   sampleKeys
//     = [NSArray arrayWithObjects: [NSNumber numberWithInt: GSKrabPlay],
//                [NSNumber numberWithInt: GSKrabStop],
//                [NSNumber numberWithInt: GSKrabEject],
//                [NSNumber numberWithInt: GSKrabTrackPrevious],
//                [NSNumber numberWithInt: GSKrabTrackNext],
//                nil];
//   dicts[0] = [NSMutableDictionary new];
//   [dicts[0] setObject: @"Cynthiune.app" forKey: @"Application"];
//   [dicts[0] setObject: @"CynthiuneClient" forKey: @"ClientName"];
//   [dicts[0] setObject: sampleKeys forKey: @"Keys"];
//   [dicts[0] setObject: @"Application" forKey: @"BindingType"];
//   dicts[1] = [NSMutableDictionary new];
//   [dicts[1] setObject: @"firefox" forKey: @"Command"];
//   [dicts[1] setObject:
//           [NSArray arrayWithObject: [NSNumber numberWithInt: GSKrabWWW]]
//         forKey: @"Keys"];
//   [dicts[1] setObject: @"ShellCommand" forKey: @"BindingType"];
//   dicts[2] = nil;

//   return dicts;
// }

// - (void) _registerKnownKeys
// {
//   NSDictionary **applicationsDicts;
//   NSDictionary **current;
//   NSArray *keys;
//   unsigned int count, max;

//   applicationsDicts = [self _knownKeys];
//   current = applicationsDicts;
//   while (*current)
//     {
//       keys = [*current objectForKey: @"Keys"];
//       max = [keys count];
//       for (count = 0; count < max; count++)
//         [self registerKey: [[keys objectAtIndex: count] intValue]
//               forObject: *current];
//       current++;
//     }
// }

- (NSConnection *) _setupConnection
{
  NSConnection *connection;
  NSPort *port;

  port = [NSPort new];
  connection = [NSConnection connectionWithReceivePort: port
                             sendPort: nil];
  if ([[NSPortNameServer systemDefaultPortNameServer] registerPort: port
                                                      forName: @"GSKrabServer"])
    {
      [connection setRootObject: server];
      [connection setReplyTimeout: 0.5];
      [connection retain];
    }
  else
    connection = nil;

  [port release];

  return connection;
}

- (id) init
{
  if ((self = [super init]))
    {
      interface = [DesktopInterface sharedInterface];
      [interface setManager: self];
      keyTable = NSCreateMapTable (NSIntMapKeyCallBacks,
                                   NSObjectMapValueCallBacks,
                                   GSKrabMaxNumberOfKeys);
//       [self _registerKnownKeys];
      server = [GSKrabServer new];
      [server setManager: self];
      serverConnection = [self _setupConnection];
      applications = [NSMutableArray new];
      if (serverConnection)
        {
          [interface start];
//           [nc addObserver: self
//               selector: @selector (connectionDied:)
//               name: NSConnectionDidDieNotification
//               object: serverConnection];
        }
      else
        {
          [self release];
          self = nil;
        }
    }

  return self;
}

- (void) dealloc
{
  [serverConnection release];
  [interface release];
  [server release];
  [applications release];
  [super dealloc];
}

- (void) registerKey: (GSKrabKeyCode) krabKeyCode
           forObject: (NSObject *) object
{
  NSMapInsert (keyTable, (void *) krabKeyCode, object);
  [interface registerKeyCode: krabKeyCode];
}

- (BOOL) _launchApplication: (NSString *) application
{
  NSWorkspace *ws;

//   NSLog (@"launching: %@", application);
  ws = [NSWorkspace sharedWorkspace];
  return [ws launchApplication: application];
}

- (BOOL) _launchCommand: (NSString *) command
{
  NSTask *task;

  task = nil;
  NS_DURING
    {
#ifdef __WIN32__
      task = [NSTask launchedTaskWithLaunchPath: @"cmd.exe"
                     arguments: [NSArray arrayWithObjects: @"start test.exe",
                                         nil]];
#else
      task = [NSTask launchedTaskWithLaunchPath: command
                     arguments: nil];
#endif
    }
  NS_HANDLER
    {
      if (![[localException name] isEqualToString: NSInvalidArgumentException])
        [localException raise];
    }
  NS_ENDHANDLER

  return (task != nil
          && ([task isRunning] || [task terminationStatus] == 0));
}

- (BOOL) _handlePendingKeyPress: (GSKrabKeyCode) krabKeyCode
                 forApplication: (NSDictionary *) application
{
#warning FIXME: should implement method
//   NSLog (@"application '%@' not launched?", application);

  return NO;
}

- (BOOL) _sendKeyPress: (GSKrabKeyCode) krabKeyCode
         toService: (NSDistantObject <GSKrabClient> *) service
{
  NSConnection *connection;
  BOOL result;

  result = NO;
  NS_DURING
    {
      connection = [service connectionForProxy];
      [connection setReplyTimeout: 0.5];
      [connection setRequestTimeout: 0.5];
      [service performActionForKey: krabKeyCode];
      result = YES;
    }
  NS_HANDLER
    {
      if (![[localException name] isEqualToString: NSPortTimeoutException])
        [localException raise];
    }
  NS_ENDHANDLER

  return result;
}

- (BOOL) _sendKeyPress: (GSKrabKeyCode) krabKeyCode
         toApplication: (NSDictionary *) application
{
  NSDistantObject <GSKrabClient> *applService;
  NSString *serviceName;
  BOOL result;

  result = NO;
  serviceName = [application objectForKey: @"ServiceName"];
  if (serviceName)
    {
      applService = [NSConnection rootProxyForConnectionWithRegisteredName: serviceName
                                  host: nil];
      if (applService)
        {
          [applService setProtocolForProxy: @protocol (GSKrabClient)];
          result = [self _sendKeyPress: krabKeyCode toService: applService];
        }
      else
        {
//           NSLog (@"service '%@' not reached", serviceName);
          result = [self _handlePendingKeyPress: krabKeyCode
                         forApplication: application];
        }
    }
  else
    {
//       NSLog (@"no service name");
      result = [self _handlePendingKeyPress: krabKeyCode
                     forApplication: application];
    }

  return result;
}

- (BOOL) handleKeyPress: (GSKrabKeyCode) krabKeyCode
{
  NSDictionary *keyHandler;
  BOOL result;

  result = NO;

  keyHandler = NSMapGet (keyTable, (void *) krabKeyCode);
  if (keyHandler)
    {
      if ([[keyHandler objectForKey: @"BindingType"]
            isEqualToString: @"Application"])
        result = [self _sendKeyPress: krabKeyCode toApplication: keyHandler];
      else if ([[keyHandler objectForKey: @"BindingType"]
                isEqualToString: @"ShellCommand"])
        result = [self _launchCommand:
                         [keyHandler objectForKey: @"Command"]];
    }
//   else
//     NSLog (@"no handler for key '%d'", krabKeyCode);

//   NSLog (@"result: %d", result);

  return result;
}

- (NSDictionary *) _findApplication: (NSString *) application
{
  NSEnumerator *enumerator;
  NSDictionary *currentDict, *applDict;

  applDict = nil;
  enumerator = [applications objectEnumerator];
  currentDict = [enumerator nextObject];
  while (currentDict && !applDict)
    if ([[currentDict objectForKey: @"Application"]
          isEqualToString: application])
      applDict = currentDict;
    else
      currentDict = [enumerator nextObject];

  return applDict;
}

- (NSDictionary *) _findApplicationWithName: (NSString *) serviceName
{
  NSEnumerator *enumerator;
  NSDictionary *currentDict, *applDict;

  applDict = nil;
  enumerator = [applications objectEnumerator];
  currentDict = [enumerator nextObject];
  while (currentDict && !applDict)
    if ([[currentDict objectForKey: @"ServiceName"]
          isEqualToString: serviceName])
      applDict = currentDict;
    else
      currentDict = [enumerator nextObject];

  return applDict;
}

- (void) registerApplication: (NSString *) application
             withServiceName: (NSString *) serviceName
{
  NSDictionary *oldApplDict;
  NSMutableDictionary *newApplDict;

  oldApplDict = [self _findApplication: application];
  if (oldApplDict)
    [applications removeObject: oldApplDict];

  newApplDict = [NSMutableDictionary new];
  [newApplDict setObject: application forKey: @"Application"];
  [newApplDict setObject: @"Application" forKey: @"BindingType"];
  [newApplDict setObject: serviceName forKey: @"ServiceName"];
  [newApplDict setObject: [NSMutableArray new] forKey: @"Keys"];
  [applications addObject: newApplDict];

//   NSLog (@"application '%@' registered with service name '%@'", serviceName);
}

- (void) applicationWithServiceName: (NSString *) serviceName
                subscribesToKeyCode: (GSKrabKeyCode) krabKeyCode
{
  NSDictionary *applDict;
  NSMutableArray *keys;
  NSNumber *keyCode;

  applDict = [self _findApplicationWithName: serviceName];
  if (applDict)
    {
      keys = [applDict objectForKey: @"Keys"];
      keyCode = [NSNumber numberWithInt: krabKeyCode];
      if ([keys indexOfObject: keyCode] == NSNotFound)
        {
          [keys addObject: keyCode];
          [self registerKey: krabKeyCode forObject: applDict];
        }
    }
//   else
//     NSLog (@"no application found with service name '%@'", serviceName);
}

- (void) stop
{
  [interface stop];
  [self release];
  exit (0);
}

@end
